import json
from typing import List, Union

import mlx.core as mx

from mflux.models.common.tokenizer import Tokenizer
from mflux.models.fibo.model.fibo_text_encoder.smol_lm3_3b_text_encoder import SmolLM3_3B_TextEncoder


class PromptEncoder:
    @staticmethod
    def encode_prompt(
        prompt: str,
        negative_prompt: str | None,
        tokenizer: Tokenizer,
        text_encoder: SmolLM3_3B_TextEncoder,
    ) -> tuple[str, mx.array, List[mx.array]]:
        # 0. Set default negative prompt if not provided
        if negative_prompt is None or negative_prompt == "":
            negative_prompt = "ugly, blurry, low quality"

        # 1. Convert prompt to JSON format
        json.loads(prompt)
        json_prompt = prompt

        # 2. Get prompt embeddings for positive and negative prompt
        prompt_embeds, prompt_layers, prompt_attention_mask = PromptEncoder._get_prompt_embeds(
            prompt=json_prompt,
            tokenizer=tokenizer,
            text_encoder=text_encoder,
            num_images_per_prompt=1,
            max_sequence_length=2048,
            tokenization_prefix="positive",
        )
        neg_prompt_embeds, neg_prompt_layers, neg_prompt_attention_mask = PromptEncoder._get_prompt_embeds(
            prompt=negative_prompt,
            tokenizer=tokenizer,
            text_encoder=text_encoder,
            num_images_per_prompt=1,
            max_sequence_length=2048,
            tokenization_prefix="negative",
        )
        encoder_hidden_states, max_tokens = PromptEncoder._get_encoder_hidden_states(
            prompt_embeds=prompt_embeds,
            neg_prompt_embeds=neg_prompt_embeds,
            prompt_attention_mask=prompt_attention_mask,
            neg_prompt_attention_mask=neg_prompt_attention_mask,
        )

        prompt_layers = PromptEncoder._get_prompt_layers(
            max_tokens=max_tokens,
            prompt_layers=prompt_layers,
            neg_prompt_layers=neg_prompt_layers,
        )
        return json_prompt, encoder_hidden_states, prompt_layers

    @staticmethod
    def _get_encoder_hidden_states(neg_prompt_attention_mask, neg_prompt_embeds, prompt_attention_mask, prompt_embeds):
        max_tokens = max(neg_prompt_embeds.shape[1], prompt_embeds.shape[1])
        prompt_embeds, prompt_attention_mask = PromptEncoder._pad_embedding(
            max_tokens=max_tokens,
            prompt_embeds=prompt_embeds,
            attention_mask=prompt_attention_mask,
        )
        neg_prompt_embeds, neg_prompt_attention_mask = PromptEncoder._pad_embedding(
            max_tokens=max_tokens,
            prompt_embeds=neg_prompt_embeds,
            attention_mask=neg_prompt_attention_mask,
        )
        encoder_hidden_states = mx.concatenate([neg_prompt_embeds, prompt_embeds], axis=0)
        return encoder_hidden_states, max_tokens

    @staticmethod
    def _get_prompt_layers(max_tokens, neg_prompt_layers, prompt_layers):
        prompt_layers = [PromptEncoder._pad_embedding(layer, max_tokens)[0] for layer in prompt_layers]
        neg_prompt_layers = [PromptEncoder._pad_embedding(layer, max_tokens)[0] for layer in neg_prompt_layers]
        prompt_layers = [mx.concatenate([neg_prompt_layers[i], prompt_layers[i]], axis=0) for i in range(len(prompt_layers))]  # fmt: off

        total_num_layers_transformer = 46  # FIBO transformer has 8 joint + 38 single blocks = 46 total layers
        if len(prompt_layers) >= total_num_layers_transformer:
            prompt_layers = prompt_layers[len(prompt_layers) - total_num_layers_transformer :]
        else:
            prompt_layers = prompt_layers + [prompt_layers[-1]] * (total_num_layers_transformer - len(prompt_layers))
        return prompt_layers

    @staticmethod
    def _get_prompt_embeds(
        prompt: Union[str, List[str]],
        text_encoder: SmolLM3_3B_TextEncoder,
        tokenizer: Tokenizer,
        num_images_per_prompt: int = 1,
        max_sequence_length: int = 2048,
        tokenization_prefix: str | None = None,
    ) -> tuple[mx.array, List[mx.array], mx.array]:
        prompts = [prompt] if isinstance(prompt, str) else list(prompt)
        if not prompts:
            raise ValueError("`prompt` must be a non-empty string or list of strings.")

        # 1) Tokenize and convert to MX
        tokenizer_output = tokenizer.tokenize(prompt=prompts, max_length=max_sequence_length)
        input_ids_mx = tokenizer_output.input_ids
        attention_mask_mx = tokenizer_output.attention_mask

        # 2) Run MLX text encoder and collect hidden states
        hidden_states_list = text_encoder(
            input_ids=input_ids_mx,
            attention_mask=attention_mask_mx,
            output_hidden_states=True,
        )

        # 3) Build prompt_embeds from last two layers
        last_hidden = hidden_states_list[-1]
        second_last_hidden = hidden_states_list[-2]
        prompt_embeds = mx.concatenate(
            [last_hidden, second_last_hidden],
            axis=-1,
        )

        # 4) Repeat along batch dimension for num_images_per_prompt
        prompt_embeds = PromptEncoder._repeat_interleave_batch(prompt_embeds, num_images_per_prompt)
        attention_mask = PromptEncoder._repeat_interleave_batch(attention_mask_mx, num_images_per_prompt)
        prompt_layers = [PromptEncoder._repeat_interleave_batch(layer, num_images_per_prompt) for layer in hidden_states_list]  # fmt: off
        return prompt_embeds, prompt_layers, attention_mask

    @staticmethod
    def _pad_embedding(
        prompt_embeds: mx.array,
        max_tokens: int,
        attention_mask: mx.array | None = None,
    ) -> tuple[mx.array, mx.array]:
        batch_size, seq_len, dim = prompt_embeds.shape

        if attention_mask is None:
            attention_mask = mx.ones((batch_size, seq_len), dtype=prompt_embeds.dtype)
        else:
            attention_mask = attention_mask.astype(prompt_embeds.dtype)

        if max_tokens < seq_len:
            raise ValueError("`max_tokens` must be >= current sequence length.")

        if max_tokens > seq_len:
            pad_length = max_tokens - seq_len
            padding = mx.zeros((batch_size, pad_length, dim), dtype=prompt_embeds.dtype)
            prompt_embeds = mx.concatenate([prompt_embeds, padding], axis=1)

            mask_padding = mx.zeros((batch_size, pad_length), dtype=attention_mask.dtype)
            attention_mask = mx.concatenate([attention_mask, mask_padding], axis=1)

        return prompt_embeds, attention_mask

    @staticmethod
    def _repeat_interleave_batch(
        tensor: mx.array,
        num_images_per_prompt: int,
    ) -> mx.array:
        if num_images_per_prompt == 1:
            return tensor
        batch, *rest = tensor.shape
        tensor = mx.expand_dims(tensor, axis=1)
        tensor = mx.broadcast_to(tensor, (batch, num_images_per_prompt, *rest))
        new_shape = (batch * num_images_per_prompt, *rest)
        return mx.reshape(tensor, new_shape)



import mlx.core.random as random

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.flux.variants.dreambooth.dataset.dataset import Dataset
from mflux.models.flux.variants.dreambooth.dataset.iterator import Iterator
from mflux.models.flux.variants.dreambooth.lora_layers.lora_layers import LoRALayers
from mflux.models.flux.variants.dreambooth.optimization.optimizer import Optimizer
from mflux.models.flux.variants.dreambooth.state.training_spec import TrainingSpec
from mflux.models.flux.variants.dreambooth.state.training_state import TrainingState
from mflux.models.flux.variants.dreambooth.statistics.statistics import Statistics
from mflux.models.flux.variants.txt2img.flux import Flux1


class DreamBoothInitializer:
    @staticmethod
    def initialize(
        config_path: str | None,
        checkpoint_path: str | None,
    ) -> tuple[Flux1, Config, TrainingSpec, TrainingState]:
        # The training specification describing the details of the training process. It is resolved
        # differently depending on if training starts from scratch or resumes from checkpoint.
        training_spec = TrainingSpec.resolve(
            config_path=config_path,
            checkpoint_path=checkpoint_path,
        )

        # Set global random seed to make training deterministic
        random.seed(training_spec.seed)

        # Load the model
        model_config = ModelConfig.from_name(training_spec.model)
        flux = Flux1(
            model_config=model_config,
            quantize=training_spec.quantize,
        )
        config = Config(
            model_config=model_config,
            num_inference_steps=training_spec.steps,
            width=training_spec.width,
            height=training_spec.height,
            guidance=training_spec.guidance,
        )

        # Create the optimizer
        optimizer = Optimizer.from_spec(training_spec)

        # Create and apply the LoRA layers directly to the transformer
        LoRALayers.from_spec(flux=flux, training_spec=training_spec)

        # Prepare the fine-tuning dataset and create the iterator
        dataset = Dataset.prepare_dataset(
            flux=flux,
            raw_data=training_spec.examples,
            width=training_spec.width,
            height=training_spec.height,
        )
        iterator = Iterator.from_spec(
            training_spec=training_spec,
            dataset=dataset,
        )

        # Setup loss statistics
        statistics = Statistics.from_spec(training_spec=training_spec)

        # The training state consisting of everything that moves during training
        training_state = TrainingState(
            optimizer=optimizer,
            iterator=iterator,
            statistics=statistics,
        )

        return flux, config, training_spec, training_state



import mlx.core as mx
from mlx import nn

from mflux.models.qwen.model.qwen_vae.qwen_image_res_block_3d import QwenImageResBlock3D
from mflux.models.qwen.model.qwen_vae.qwen_image_resample_3d import QwenImageResample3D


class QwenImageDownBlock3D(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, num_res_blocks: int = 2, downsample_mode: str = None):
        super().__init__()
        self.resnets = []
        current_dim = in_channels
        for _ in range(num_res_blocks):
            self.resnets.append(QwenImageResBlock3D(current_dim, out_channels))
            current_dim = out_channels

        self.downsamplers = None
        if downsample_mode is not None:
            self.downsamplers = [QwenImageResample3D(out_channels, mode=downsample_mode)]

    def __call__(self, x: mx.array) -> mx.array:
        for resnet in self.resnets:
            x = resnet(x)
        if self.downsamplers is not None:
            x = self.downsamplers[0](x)
        return x



from mflux.models.flux.weights.flux_weight_definition import FluxWeightDefinition
from mflux.models.flux.weights.flux_weight_mapping import FluxWeightMapping

__all__ = ["FluxWeightDefinition", "FluxWeightMapping"]



from dataclasses import dataclass


@dataclass
class MetaData:
    quantization_level: int | None = None
    mflux_version: str | None = None


@dataclass
class LoadedWeights:
    components: dict[str, dict]
    meta_data: MetaData

    def __getattr__(self, name: str) -> dict | None:
        if name in ("components", "meta_data"):
            return object.__getattribute__(self, name)
        return self.components.get(name)

    def num_transformer_blocks(self, component_name: str = "transformer") -> int:
        transformer = self.components.get(component_name)
        if transformer is None:
            for comp in self.components.values():
                if isinstance(comp, dict) and "transformer_blocks" in comp:
                    transformer = comp
                    break
        if transformer and "transformer_blocks" in transformer:
            return len(transformer["transformer_blocks"])
        return 0

    def num_single_transformer_blocks(self, component_name: str = "transformer") -> int:
        transformer = self.components.get(component_name)
        if transformer is None:
            for comp in self.components.values():
                if isinstance(comp, dict) and "single_transformer_blocks" in comp:
                    transformer = comp
                    break
        if transformer and "single_transformer_blocks" in transformer:
            return len(transformer["single_transformer_blocks"])
        return 0



import mlx.core as mx
from mlx import nn


class Wan2_2_AvgDown3D(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        factor_t: int,
        factor_s: int = 1,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.factor_t = factor_t
        self.factor_s = factor_s
        self.factor = self.factor_t * self.factor_s * self.factor_s

        assert in_channels * self.factor % out_channels == 0
        self.group_size = in_channels * self.factor // out_channels

    def __call__(self, x: mx.array) -> mx.array:
        pad_t = (self.factor_t - x.shape[2] % self.factor_t) % self.factor_t
        if pad_t > 0:
            x = mx.pad(
                x,
                [
                    (0, 0),  # batch
                    (0, 0),  # channels
                    (pad_t, 0),  # time
                    (0, 0),  # height
                    (0, 0),  # width
                ],
            )

        b, c, t, h, w = x.shape

        x = mx.reshape(
            x,
            (
                b,
                c,
                t // self.factor_t,
                self.factor_t,
                h // self.factor_s,
                self.factor_s,
                w // self.factor_s,
                self.factor_s,
            ),
        )
        x = mx.transpose(x, (0, 1, 3, 5, 7, 2, 4, 6))
        x = mx.reshape(
            x,
            (
                b,
                c * self.factor,
                t // self.factor_t,
                h // self.factor_s,
                w // self.factor_s,
            ),
        )
        x = mx.reshape(
            x,
            (
                b,
                self.out_channels,
                self.group_size,
                t // self.factor_t,
                h // self.factor_s,
                w // self.factor_s,
            ),
        )
        x = mx.mean(x, axis=2)
        return x



import mlx.core as mx
from mlx import nn

from mflux.models.qwen.model.qwen_text_encoder.qwen_encoder import QwenEncoder


class QwenTextEncoder(nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = QwenEncoder()

    def __call__(
        self,
        input_ids: mx.array,
        attention_mask: mx.array,
    ) -> tuple[mx.array, mx.array]:
        hidden_states = self.encoder(input_ids, attention_mask)

        prompt_embeds, encoder_attention_mask = QwenTextEncoder._process_text_embeddings_mlx(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            drop_idx=34,
            dtype=mx.bfloat16,
        )

        return prompt_embeds, encoder_attention_mask

    @staticmethod
    def _process_text_embeddings_mlx(hidden_states, attention_mask, drop_idx=1, dtype=mx.float32):
        split_hidden_states = QwenTextEncoder._extract_masked_hidden(hidden_states, attention_mask)
        split_hidden_states = [e[drop_idx:] for e in split_hidden_states]
        attn_mask_list = [mx.ones(e.shape[0], dtype=mx.int32) for e in split_hidden_states]
        max_seq_len = max([e.shape[0] for e in split_hidden_states])

        padded_embeds = []
        for u in split_hidden_states:
            current_len = u.shape[0]
            hidden_dim = u.shape[1]
            if current_len < max_seq_len:
                padding = mx.zeros((max_seq_len - current_len, hidden_dim), dtype=u.dtype)
                padded = mx.concatenate([u, padding], axis=0)
            else:
                padded = u
            padded_embeds.append(padded)

        prompt_embeds = mx.stack(padded_embeds, axis=0)

        padded_masks = []
        for mask in attn_mask_list:
            current_len = mask.shape[0]
            if current_len < max_seq_len:
                padding = mx.zeros(max_seq_len - current_len, dtype=mask.dtype)
                padded = mx.concatenate([mask, padding], axis=0)
            else:
                padded = mask
            padded_masks.append(padded)

        encoder_attention_mask = mx.stack(padded_masks, axis=0)
        prompt_embeds = prompt_embeds.astype(dtype)
        return prompt_embeds, encoder_attention_mask

    @staticmethod
    def _extract_masked_hidden(hidden_states, attention_mask):
        batch_size = hidden_states.shape[0]
        split_hidden_states = []
        for i in range(batch_size):
            mask = attention_mask[i]
            valid_length = mx.sum(mask).item()
            valid_length = int(valid_length)
            valid_hidden = hidden_states[i, :valid_length, :]
            split_hidden_states.append(valid_hidden)
        return split_hidden_states



import math

import mlx.core as mx
from mlx import nn
from mlx.core.fast import scaled_dot_product_attention

from mflux.models.fibo_vlm.model.qwen3_vl_rms_norm import Qwen3VLRMSNorm


class Qwen3VLAttention(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        num_attention_heads: int,
        num_key_value_heads: int,
        head_dim: int,
        max_position_embeddings: int = 262144,
        rope_theta: float = 1000000.0,
        mrope_section: list[int] | None = None,
        attention_bias: bool = False,
        rms_norm_eps: float = 1e-6,
    ):
        super().__init__()
        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads
        self.num_key_value_heads = num_key_value_heads
        self.head_dim = head_dim
        self.num_key_value_groups = num_attention_heads // num_key_value_heads
        self.scaling = 1.0 / math.sqrt(self.head_dim)
        self.q_proj = nn.Linear(hidden_size, num_attention_heads * head_dim, bias=attention_bias)
        self.k_proj = nn.Linear(hidden_size, num_key_value_heads * head_dim, bias=attention_bias)
        self.v_proj = nn.Linear(hidden_size, num_key_value_heads * head_dim, bias=attention_bias)
        self.o_proj = nn.Linear(num_attention_heads * head_dim, hidden_size, bias=attention_bias)
        self.q_norm = Qwen3VLRMSNorm(head_dim, eps=rms_norm_eps)
        self.k_norm = Qwen3VLRMSNorm(head_dim, eps=rms_norm_eps)
        self.mrope_section = mrope_section or [24, 20, 20]

    def __call__(
        self,
        hidden_states: mx.array,
        attention_mask: mx.array | None = None,
        position_embeddings: tuple[mx.array, mx.array] | None = None,
        past_key_value: tuple[mx.array, mx.array] | None = None,
    ) -> mx.array | tuple[mx.array, tuple[mx.array, mx.array]]:
        bsz, q_len, _ = hidden_states.shape

        q_proj = self.q_proj(hidden_states)
        k_proj = self.k_proj(hidden_states)
        v_proj = self.v_proj(hidden_states)

        query_states = q_proj.reshape(bsz, q_len, self.num_attention_heads, self.head_dim)
        key_states = k_proj.reshape(bsz, q_len, self.num_key_value_heads, self.head_dim)
        value_states = v_proj.reshape(bsz, q_len, self.num_key_value_heads, self.head_dim)

        query_states = self.q_norm(query_states)
        key_states = self.k_norm(key_states)

        query_states = query_states.transpose(0, 2, 1, 3)
        key_states = key_states.transpose(0, 2, 1, 3)
        value_states = value_states.transpose(0, 2, 1, 3)

        if position_embeddings is not None:
            cos, sin = position_embeddings
            query_states, key_states = Qwen3VLAttention._apply_rotary_pos_emb(
                q=query_states,
                k=key_states,
                cos=cos,
                sin=sin,
            )

        cache_key_states = key_states
        cache_value_states = value_states

        if past_key_value is not None:
            past_key, past_value = past_key_value
            cache_key_states = mx.concatenate([past_key, cache_key_states], axis=2)
            cache_value_states = mx.concatenate([past_value, cache_value_states], axis=2)

        if self.num_key_value_heads != self.num_attention_heads:
            key_states = Qwen3VLAttention._repeat_kv(cache_key_states, self.num_key_value_groups)
            value_states = Qwen3VLAttention._repeat_kv(cache_value_states, self.num_key_value_groups)
        else:
            key_states = cache_key_states
            value_states = cache_value_states

        attn_mask = None
        if attention_mask is not None:
            kv_len = key_states.shape[2]
            attn_mask = attention_mask[:, :, :, :kv_len]

        query_states_f32 = query_states.astype(mx.float32)
        key_states_f32 = key_states.astype(mx.float32)
        value_states_f32 = value_states.astype(mx.float32)

        # Use fast MLX attention
        attn_output = scaled_dot_product_attention(
            query_states_f32,
            key_states_f32,
            value_states_f32,
            scale=self.scaling,
            mask=attn_mask,
        )

        attn_output = attn_output.astype(query_states.dtype)
        attn_output = attn_output.transpose(0, 2, 1, 3).reshape(bsz, q_len, self.num_attention_heads * self.head_dim)
        attn_output = self.o_proj(attn_output)
        return attn_output, (cache_key_states, cache_value_states)

    @staticmethod
    def _repeat_kv(hidden_states: mx.array, n_rep: int) -> mx.array:
        shape = hidden_states.shape

        if len(shape) == 5:
            batch, num_key_value_heads, rep, slen, head_dim = shape
            return hidden_states.reshape(batch, num_key_value_heads * rep, slen, head_dim)

        batch, num_key_value_heads, slen, head_dim = shape
        hidden_states = mx.expand_dims(hidden_states, axis=2)
        hidden_states = mx.broadcast_to(hidden_states, (batch, num_key_value_heads, n_rep, slen, head_dim))
        return hidden_states.reshape(batch, num_key_value_heads * n_rep, slen, head_dim)

    @staticmethod
    def _apply_rotary_pos_emb(
        q: mx.array,
        k: mx.array,
        cos: mx.array,
        sin: mx.array,
        unsqueeze_dim: int = 1,
    ) -> tuple[mx.array, mx.array]:
        cos = mx.expand_dims(cos, axis=unsqueeze_dim)
        sin = mx.expand_dims(sin, axis=unsqueeze_dim)
        q_embed = (q * cos) + (Qwen3VLAttention._rotate_half(q) * sin)
        k_embed = (k * cos) + (Qwen3VLAttention._rotate_half(k) * sin)
        return q_embed, k_embed

    @staticmethod
    def _rotate_half(x: mx.array) -> mx.array:
        x1 = x[..., : x.shape[-1] // 2]
        x2 = x[..., x.shape[-1] // 2 :]
        return mx.concatenate([-x2, x1], axis=-1)



import mlx.core as mx
from mlx import nn


class FiboGELU(nn.Module):
    def __init__(self, dim_in: int, dim_out: int, approximate: str = "tanh", bias: bool = True):
        super().__init__()
        self.proj = nn.Linear(dim_in, dim_out, bias=bias)
        self.approximate = approximate

    def __call__(self, hidden_states: mx.array) -> mx.array:
        hidden_states = self.proj(hidden_states)
        hidden_states = nn.gelu_approx(hidden_states)
        return hidden_states



from typing import List

from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.tokenizer import LanguageTokenizer
from mflux.models.common.weights.loading.weight_definition import ComponentDefinition, TokenizerDefinition
from mflux.models.z_image.weights.z_image_weight_mapping import ZImageWeightMapping


class ZImageWeightDefinition:
    @staticmethod
    def get_components() -> List[ComponentDefinition]:
        return [
            ComponentDefinition(
                name="vae",
                hf_subdir="vae",
                num_blocks=4,
                precision=ModelConfig.precision,
                mapping_getter=ZImageWeightMapping.get_vae_mapping,
            ),
            ComponentDefinition(
                name="transformer",
                hf_subdir="transformer",
                num_layers=30,
                precision=ModelConfig.precision,
                mapping_getter=ZImageWeightMapping.get_transformer_mapping,
            ),
            ComponentDefinition(
                name="text_encoder",
                hf_subdir="text_encoder",
                num_layers=36,
                precision=ModelConfig.precision,
                mapping_getter=ZImageWeightMapping.get_text_encoder_mapping,
            ),
        ]

    @staticmethod
    def get_tokenizers() -> List[TokenizerDefinition]:
        return [
            TokenizerDefinition(
                name="z_image",
                hf_subdir="tokenizer",
                tokenizer_class="AutoTokenizer",
                encoder_class=LanguageTokenizer,
                max_length=512,
                use_chat_template=True,
                chat_template_kwargs={"enable_thinking": True},
                download_patterns=["tokenizer/*"],
            ),
        ]

    @staticmethod
    def get_download_patterns() -> List[str]:
        return [
            "vae/*.safetensors",
            "vae/*.json",
            "transformer/*.safetensors",
            "transformer/*.json",
            "text_encoder/*.safetensors",
            "text_encoder/*.json",
            "tokenizer/*",
        ]

    @staticmethod
    def quantization_predicate(path: str, module) -> bool:
        return hasattr(module, "to_quantized")



import mlx.core as mx
from mlx import nn
from mlx.core.fast import scaled_dot_product_attention


class ZImageAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        n_heads: int,
        qk_norm: bool = True,
        eps: float = 1e-5,
    ):
        super().__init__()
        self.dim = dim
        self.n_heads = n_heads
        self.head_dim = dim // n_heads
        self.scale = self.head_dim**-0.5

        # Projections (no bias)
        self.to_q = nn.Linear(dim, dim, bias=False)
        self.to_k = nn.Linear(dim, dim, bias=False)
        self.to_v = nn.Linear(dim, dim, bias=False)
        self.to_out = [nn.Linear(dim, dim, bias=False)]

        # Optional QK normalization
        if qk_norm:
            self.norm_q = nn.RMSNorm(self.head_dim, eps=eps)
            self.norm_k = nn.RMSNorm(self.head_dim, eps=eps)
        else:
            self.norm_q = None
            self.norm_k = None

    def __call__(
        self,
        hidden_states: mx.array,
        attention_mask: mx.array | None = None,
        freqs_cis: mx.array | None = None,
    ) -> mx.array:
        batch_size, seq_len, _ = hidden_states.shape

        # Project to Q, K, V
        query = self.to_q(hidden_states)
        key = self.to_k(hidden_states)
        value = self.to_v(hidden_states)

        # Reshape to (batch, seq_len, heads, head_dim)
        query = query.reshape(batch_size, seq_len, self.n_heads, self.head_dim)
        key = key.reshape(batch_size, seq_len, self.n_heads, self.head_dim)
        value = value.reshape(batch_size, seq_len, self.n_heads, self.head_dim)

        # Apply QK normalization
        if self.norm_q is not None:
            query = self.norm_q(query)
        if self.norm_k is not None:
            key = self.norm_k(key)

        # Apply RoPE
        if freqs_cis is not None:
            query = ZImageAttention._apply_rotary_emb(query, freqs_cis)
            key = ZImageAttention._apply_rotary_emb(key, freqs_cis)

        # Transpose for attention: (batch, heads, seq_len, head_dim)
        query = mx.transpose(query, axes=(0, 2, 1, 3))
        key = mx.transpose(key, axes=(0, 2, 1, 3))
        value = mx.transpose(value, axes=(0, 2, 1, 3))

        # Convert boolean mask to additive mask for SDPA
        mask = None
        if attention_mask is not None:
            mask = mx.where(attention_mask[:, None, None, :], mx.array(0.0), mx.array(float("-inf")))

        hidden_states = scaled_dot_product_attention(query, key, value, scale=self.scale, mask=mask)
        hidden_states = mx.transpose(hidden_states, axes=(0, 2, 1, 3))
        hidden_states = hidden_states.reshape(batch_size, seq_len, self.dim)
        hidden_states = self.to_out[0](hidden_states)
        return hidden_states

    @staticmethod
    def _apply_rotary_emb(x: mx.array, freqs_cis: mx.array) -> mx.array:
        batch_size, seq_len, n_heads, head_dim = x.shape
        x = x.reshape(batch_size, seq_len, n_heads, head_dim // 2, 2)
        freqs_cis = mx.expand_dims(freqs_cis, axis=0)
        freqs_cis = mx.expand_dims(freqs_cis, axis=2)
        x_real, x_imag = x[..., 0], x[..., 1]
        freqs_cos, freqs_sin = freqs_cis[..., 0], freqs_cis[..., 1]
        x_out_real = x_real * freqs_cos - x_imag * freqs_sin
        x_out_imag = x_real * freqs_sin + x_imag * freqs_cos
        x_out = mx.stack([x_out_real, x_out_imag], axis=-1)
        return x_out.reshape(batch_size, seq_len, n_heads, head_dim)



import mlx.core as mx
from mlx import nn
from mlx.core.fast import scaled_dot_product_attention


class Attention(nn.Module):
    def __init__(
        self,
        hidden_size: int = 2560,
        num_attention_heads: int = 32,
        num_key_value_heads: int = 8,
        head_dim: int = 128,
    ):
        super().__init__()
        self.num_heads = num_attention_heads
        self.num_kv_heads = num_key_value_heads
        self.head_dim = head_dim
        self.num_kv_groups = num_attention_heads // num_key_value_heads
        self.scale = head_dim**-0.5

        self.q_proj = nn.Linear(hidden_size, num_attention_heads * head_dim, bias=False)
        self.k_proj = nn.Linear(hidden_size, num_key_value_heads * head_dim, bias=False)
        self.v_proj = nn.Linear(hidden_size, num_key_value_heads * head_dim, bias=False)
        self.o_proj = nn.Linear(num_attention_heads * head_dim, hidden_size, bias=False)
        self.q_norm = nn.RMSNorm(head_dim)
        self.k_norm = nn.RMSNorm(head_dim)

    def __call__(
        self,
        hidden_states: mx.array,
        attention_mask: mx.array | None = None,
        position_embeddings: tuple[mx.array, mx.array] | None = None,
    ) -> mx.array:
        batch_size, seq_len, _ = hidden_states.shape

        q = self.q_proj(hidden_states).reshape(batch_size, seq_len, self.num_heads, self.head_dim)
        k = self.k_proj(hidden_states).reshape(batch_size, seq_len, self.num_kv_heads, self.head_dim)
        v = self.v_proj(hidden_states).reshape(batch_size, seq_len, self.num_kv_heads, self.head_dim)

        q = self.q_norm(q)
        k = self.k_norm(k)

        if position_embeddings is not None:
            q, k = Attention._apply_rotary_pos_emb(q, k, *position_embeddings)

        if self.num_kv_groups > 1:
            k = mx.repeat(k, self.num_kv_groups, axis=2)
            v = mx.repeat(v, self.num_kv_groups, axis=2)

        q = mx.transpose(q, axes=(0, 2, 1, 3))
        k = mx.transpose(k, axes=(0, 2, 1, 3))
        v = mx.transpose(v, axes=(0, 2, 1, 3))

        attn_output = scaled_dot_product_attention(q, k, v, scale=self.scale, mask=attention_mask)
        attn_output = mx.transpose(attn_output, axes=(0, 2, 1, 3)).reshape(batch_size, seq_len, -1)
        return self.o_proj(attn_output)

    @staticmethod
    def _apply_rotary_pos_emb(
        q: mx.array,
        k: mx.array,
        cos: mx.array,
        sin: mx.array,
        unsqueeze_dim: int = 2,
    ) -> tuple[mx.array, mx.array]:
        cos = mx.expand_dims(cos, axis=unsqueeze_dim)
        sin = mx.expand_dims(sin, axis=unsqueeze_dim)
        q_embed = (q * cos) + (Attention._rotate_half(q) * sin)
        k_embed = (k * cos) + (Attention._rotate_half(k) * sin)
        return q_embed, k_embed

    @staticmethod
    def _rotate_half(x: mx.array) -> mx.array:
        x1 = x[..., : x.shape[-1] // 2]
        x2 = x[..., x.shape[-1] // 2 :]
        return mx.concatenate([-x2, x1], axis=-1)



from mflux.models.common.lora.mapping.lora_mapping import LoRAMapping, LoRATarget


class ZImageLoRAMapping(LoRAMapping):
    @staticmethod
    def get_mapping() -> list[LoRATarget]:
        targets = []
        for layer_type in ["layers", "noise_refiner", "context_refiner"]:
            targets.extend(ZImageLoRAMapping._get_layer_targets(layer_type))
        targets.extend(ZImageLoRAMapping._get_global_targets())
        return targets

    @staticmethod
    def _get_global_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="all_x_embedder.2-1",
                possible_up_patterns=["diffusion_model.all_x_embedder.2-1.lora_B.weight"],
                possible_down_patterns=["diffusion_model.all_x_embedder.2-1.lora_A.weight"],
            ),
            LoRATarget(
                model_path="all_final_layer.2-1.linear",
                possible_up_patterns=["diffusion_model.all_final_layer.2-1.linear.lora_B.weight"],
                possible_down_patterns=["diffusion_model.all_final_layer.2-1.linear.lora_A.weight"],
            ),
            LoRATarget(
                model_path="all_final_layer.2-1.adaLN_modulation.0",
                possible_up_patterns=["diffusion_model.all_final_layer.2-1.adaLN_modulation.1.lora_B.weight"],
                possible_down_patterns=["diffusion_model.all_final_layer.2-1.adaLN_modulation.1.lora_A.weight"],
            ),
            LoRATarget(
                model_path="cap_embedder.1",
                possible_up_patterns=["diffusion_model.cap_embedder.1.lora_B.weight"],
                possible_down_patterns=["diffusion_model.cap_embedder.1.lora_A.weight"],
            ),
            LoRATarget(
                model_path="t_embedder.linear1",
                possible_up_patterns=["diffusion_model.t_embedder.mlp.0.lora_B.weight"],
                possible_down_patterns=["diffusion_model.t_embedder.mlp.0.lora_A.weight"],
            ),
            LoRATarget(
                model_path="t_embedder.linear2",
                possible_up_patterns=["diffusion_model.t_embedder.mlp.2.lora_B.weight"],
                possible_down_patterns=["diffusion_model.t_embedder.mlp.2.lora_A.weight"],
            ),
        ]

    @staticmethod
    def _get_layer_targets(layer_type: str) -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.adaLN_modulation.0",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.adaLN_modulation.0.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.adaLN_modulation.0.lora_up.weight",
                    f"{layer_type}.{{block}}.adaLN_modulation.0.lora_B.weight",
                    f"{layer_type}.{{block}}.adaLN_modulation.0.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_adaLN_modulation_0.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.adaLN_modulation.0.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.adaLN_modulation.0.lora_down.weight",
                    f"{layer_type}.{{block}}.adaLN_modulation.0.lora_A.weight",
                    f"{layer_type}.{{block}}.adaLN_modulation.0.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_adaLN_modulation_0.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.adaLN_modulation.0.alpha",
                    f"{layer_type}.{{block}}.adaLN_modulation.0.alpha",
                    f"lora_unet_{layer_type}_{{block}}_adaLN_modulation_0.alpha",
                ],
            ),
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.attention.to_q",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_q.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_q.lora_up.weight",
                    f"{layer_type}.{{block}}.attention.to_q.lora_B.weight",
                    f"{layer_type}.{{block}}.attention.to_q.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_q.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_q.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_q.lora_down.weight",
                    f"{layer_type}.{{block}}.attention.to_q.lora_A.weight",
                    f"{layer_type}.{{block}}.attention.to_q.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_q.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_q.alpha",
                    f"{layer_type}.{{block}}.attention.to_q.alpha",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_q.alpha",
                ],
            ),
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.attention.to_k",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_k.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_k.lora_up.weight",
                    f"{layer_type}.{{block}}.attention.to_k.lora_B.weight",
                    f"{layer_type}.{{block}}.attention.to_k.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_k.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_k.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_k.lora_down.weight",
                    f"{layer_type}.{{block}}.attention.to_k.lora_A.weight",
                    f"{layer_type}.{{block}}.attention.to_k.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_k.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_k.alpha",
                    f"{layer_type}.{{block}}.attention.to_k.alpha",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_k.alpha",
                ],
            ),
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.attention.to_v",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_v.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_v.lora_up.weight",
                    f"{layer_type}.{{block}}.attention.to_v.lora_B.weight",
                    f"{layer_type}.{{block}}.attention.to_v.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_v.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_v.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_v.lora_down.weight",
                    f"{layer_type}.{{block}}.attention.to_v.lora_A.weight",
                    f"{layer_type}.{{block}}.attention.to_v.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_v.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_v.alpha",
                    f"{layer_type}.{{block}}.attention.to_v.alpha",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_v.alpha",
                ],
            ),
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.attention.to_out.0",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_out.0.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_out.0.lora_up.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.out.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.out.lora_up.weight",
                    f"{layer_type}.{{block}}.attention.to_out.0.lora_B.weight",
                    f"{layer_type}.{{block}}.attention.to_out.0.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_out_0.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_out.0.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_out.0.lora_down.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.out.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.attention.out.lora_down.weight",
                    f"{layer_type}.{{block}}.attention.to_out.0.lora_A.weight",
                    f"{layer_type}.{{block}}.attention.to_out.0.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_out_0.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.attention.to_out.0.alpha",
                    f"diffusion_model.{layer_type}.{{block}}.attention.out.alpha",
                    f"{layer_type}.{{block}}.attention.to_out.0.alpha",
                    f"lora_unet_{layer_type}_{{block}}_attention_to_out_0.alpha",
                ],
            ),
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.feed_forward.w1",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w1.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w1.lora_up.weight",
                    f"{layer_type}.{{block}}.feed_forward.w1.lora_B.weight",
                    f"{layer_type}.{{block}}.feed_forward.w1.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w1.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w1.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w1.lora_down.weight",
                    f"{layer_type}.{{block}}.feed_forward.w1.lora_A.weight",
                    f"{layer_type}.{{block}}.feed_forward.w1.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w1.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w1.alpha",
                    f"{layer_type}.{{block}}.feed_forward.w1.alpha",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w1.alpha",
                ],
            ),
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.feed_forward.w2",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w2.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w2.lora_up.weight",
                    f"{layer_type}.{{block}}.feed_forward.w2.lora_B.weight",
                    f"{layer_type}.{{block}}.feed_forward.w2.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w2.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w2.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w2.lora_down.weight",
                    f"{layer_type}.{{block}}.feed_forward.w2.lora_A.weight",
                    f"{layer_type}.{{block}}.feed_forward.w2.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w2.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w2.alpha",
                    f"{layer_type}.{{block}}.feed_forward.w2.alpha",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w2.alpha",
                ],
            ),
            LoRATarget(
                model_path=f"{layer_type}.{{block}}.feed_forward.w3",
                possible_up_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w3.lora_B.weight",
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w3.lora_up.weight",
                    f"{layer_type}.{{block}}.feed_forward.w3.lora_B.weight",
                    f"{layer_type}.{{block}}.feed_forward.w3.lora_up.weight",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w3.lora_up.weight",
                ],
                possible_down_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w3.lora_A.weight",
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w3.lora_down.weight",
                    f"{layer_type}.{{block}}.feed_forward.w3.lora_A.weight",
                    f"{layer_type}.{{block}}.feed_forward.w3.lora_down.weight",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w3.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    f"diffusion_model.{layer_type}.{{block}}.feed_forward.w3.alpha",
                    f"{layer_type}.{{block}}.feed_forward.w3.alpha",
                    f"lora_unet_{layer_type}_{{block}}_feed_forward_w3.alpha",
                ],
            ),
        ]



import mlx.core as mx
from mlx import nn

from mflux.models.fibo_vlm.model.qwen3_vl_attention import Qwen3VLAttention
from mflux.models.fibo_vlm.model.qwen3_vl_mlp import Qwen3VLMLP
from mflux.models.fibo_vlm.model.qwen3_vl_rms_norm import Qwen3VLRMSNorm


class Qwen3VLDecoderLayer(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        num_attention_heads: int,
        num_key_value_heads: int,
        head_dim: int,
        max_position_embeddings: int,
        rope_theta: float,
        mrope_section: list[int] | None,
        attention_bias: bool,
        rms_norm_eps: float,
        intermediate_size: int,
    ):
        super().__init__()
        self.input_layernorm = Qwen3VLRMSNorm(hidden_size, eps=rms_norm_eps)
        self.self_attn = Qwen3VLAttention(
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            num_key_value_heads=num_key_value_heads,
            head_dim=head_dim,
            max_position_embeddings=max_position_embeddings,
            rope_theta=rope_theta,
            mrope_section=mrope_section,
            attention_bias=attention_bias,
            rms_norm_eps=rms_norm_eps,
        )
        self.post_attention_layernorm = Qwen3VLRMSNorm(hidden_size, eps=rms_norm_eps)
        self.mlp = Qwen3VLMLP(
            hidden_size=hidden_size,
            intermediate_size=intermediate_size,
        )

    def __call__(
        self,
        hidden_states: mx.array,
        attention_mask: mx.array | None = None,
        position_embeddings: tuple[mx.array, mx.array] | None = None,
        past_key_value: tuple[mx.array, mx.array] | None = None,
    ) -> mx.array | tuple[mx.array, tuple[mx.array, mx.array]]:
        residual = hidden_states
        hidden_states = self.input_layernorm(hidden_states)
        attn_output, past_key_value = self.self_attn(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            position_embeddings=position_embeddings,
            past_key_value=past_key_value,
        )
        hidden_states = residual + attn_output
        residual = hidden_states
        hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)
        hidden_states = residual + hidden_states
        return hidden_states, past_key_value



import mlx.core as mx
from mlx import nn


class VisionRotaryEmbedding(nn.Module):
    def __init__(self, dim: int, theta: float = 10000.0):
        super().__init__()
        self.dim = dim
        self.theta = theta
        inv_freq = 1.0 / (theta ** (mx.arange(0, dim, 2, dtype=mx.float32) / dim))
        self.inv_freq = inv_freq

    def __call__(self, max_grid_size: int) -> mx.array:
        positions = mx.arange(max_grid_size, dtype=mx.float32)
        freqs = mx.outer(positions, self.inv_freq)
        return freqs



import datetime
import json
import os
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import List

from mflux.models.flux.variants.dreambooth.state.zip_util import ZipUtil


@dataclass
class ExampleSpec:
    image: Path
    prompt: str

    @classmethod
    def create(cls, param: dict[str, str], absolute_or_relative_path: str, base_path: Path) -> "ExampleSpec":
        image_path = Path(param["image"])

        # If the path is not absolute, resolve it relative to the base path
        if not Path(absolute_or_relative_path).is_absolute():
            image_path = Path(base_path).parent / absolute_or_relative_path / image_path
        else:
            image_path = Path(absolute_or_relative_path) / image_path

        return cls(
            image=image_path,
            prompt=param["prompt"],
        )


@dataclass
class TrainingLoopSpec:
    num_epochs: int
    batch_size: int
    iterator_state_path: str | None = None


@dataclass
class OptimizerSpec:
    name: str
    learning_rate: float
    state_path: str | None = None


@dataclass
class SaveSpec:
    checkpoint_frequency: int
    output_path: str


@dataclass
class StatisticsSpec:
    state_path: str | None = None


@dataclass
class BlockRange:
    start: int | None = None
    end: int | None = None
    indices: List[int] | None = None

    def get_blocks(self) -> List[int]:
        if self.indices:
            return self.indices
        if self.start is not None and self.end is not None:
            return list(range(self.start, self.end))
        raise ValueError("Either 'start' and 'end' or 'indices' must be provided.")


@dataclass
class TransformerBlocks:
    block_range: BlockRange
    layer_types: List[str]
    lora_rank: int


@dataclass
class SingleTransformerBlocks:
    block_range: BlockRange
    layer_types: List[str]
    lora_rank: int


@dataclass
class LoraLayersSpec:
    transformer_blocks: TransformerBlocks
    single_transformer_blocks: SingleTransformerBlocks
    state_path: str | None = None


@dataclass
class InstrumentationSpec:
    plot_frequency: int
    generate_image_frequency: int
    validation_prompt: str


@dataclass
class TrainingSpec:
    model: str
    seed: int
    steps: int
    guidance: float
    quantize: int
    width: int
    height: int
    training_loop: TrainingLoopSpec
    optimizer: OptimizerSpec
    saver: SaveSpec
    instrumentation: InstrumentationSpec
    lora_layers: LoraLayersSpec
    statistics: StatisticsSpec
    examples: List[ExampleSpec]
    config_path: str | None = None
    checkpoint_path: str | None = None

    @staticmethod
    def resolve(config_path: str | None, checkpoint_path: str | None) -> "TrainingSpec":
        if not config_path and not checkpoint_path:
            raise ValueError("Either 'config_path' or 'checkpoint_path' must be provided.")

        if checkpoint_path:
            return TrainingSpec._from_checkpoint(checkpoint_path, new_folder=False)

        return TrainingSpec._from_config(config_path, new_folder=True)

    @staticmethod
    def _from_config(path: str, new_folder: bool = True) -> "TrainingSpec":
        with open(Path(path), "r") as f:
            data = json.load(f)

        return TrainingSpec.from_conf(data, path, new_folder)

    @staticmethod
    def from_conf(config: dict, config_path: str | None, new_folder: bool = True) -> "TrainingSpec":
        absolute_config_path = None if config_path is None else Path(config_path).absolute()
        absolute_or_relative_path = config["examples"]["path"]
        examples = [ExampleSpec.create(ex, absolute_or_relative_path, absolute_config_path) for ex in config["examples"]["images"]]  # fmt: off
        training_loop = TrainingLoopSpec(**config["training_loop"])
        optimizer = OptimizerSpec(**config["optimizer"])
        saver = SaveSpec(
            checkpoint_frequency=config["save"]["checkpoint_frequency"],
            output_path=TrainingSpec._resolve_output_path(config["save"]["output_path"], new_folder),
        )
        instrumentation = (
            None if config.get("instrumentation", None) is None else InstrumentationSpec(**config["instrumentation"])
        )
        statistics = (
            StatisticsSpec() if config.get("statistics", None) is None else StatisticsSpec(**config["statistics"])
        )

        # Parse LoraLayers structure
        transformer_blocks = config["lora_layers"].get("transformer_blocks", None)
        if transformer_blocks:
            transformer_blocks = TransformerBlocks(
                block_range=BlockRange(**transformer_blocks["block_range"]),
                layer_types=transformer_blocks["layer_types"],
                lora_rank=transformer_blocks["lora_rank"],
            )

        single_transformer_blocks = config["lora_layers"].get("single_transformer_blocks", None)
        if single_transformer_blocks:
            single_transformer_blocks = SingleTransformerBlocks(
                block_range=BlockRange(**single_transformer_blocks["block_range"]),
                layer_types=single_transformer_blocks["layer_types"],
                lora_rank=single_transformer_blocks["lora_rank"],
            )

        lora_layers = LoraLayersSpec(
            transformer_blocks=transformer_blocks, single_transformer_blocks=single_transformer_blocks
        )

        # Create the training specification
        return TrainingSpec(
            model=config["model"],
            seed=config["seed"],
            steps=config["steps"],
            guidance=config["guidance"],
            quantize=config["quantize"],
            width=config["width"],
            height=config["height"],
            training_loop=training_loop,
            optimizer=optimizer,
            saver=saver,
            instrumentation=instrumentation,
            lora_layers=lora_layers,
            statistics=statistics,
            examples=examples,
            config_path=None if absolute_config_path is None else str(absolute_config_path),
        )

    def to_json(self) -> str:
        spec_dict = asdict(self)
        serialized_dict = TrainingSpec._custom_serializer(spec_dict)
        return json.dumps(serialized_dict, indent=4)

    @staticmethod
    def _custom_serializer(obj):
        if isinstance(obj, Path):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(obj, list):
            return [TrainingSpec._custom_serializer(item) for item in obj]
        elif isinstance(obj, dict):
            return {key: TrainingSpec._custom_serializer(value) for key, value in obj.items()}
        return obj

    @staticmethod
    def _resolve_output_path(path: str, new_folder: bool) -> str:
        requested_path = os.path.expanduser(path)
        now_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

        if new_folder and os.path.exists(requested_path):
            requested_path = f"{requested_path}_{now_str}"

        os.makedirs(requested_path, exist_ok=True)
        return requested_path

    @staticmethod
    def _from_checkpoint(path: str, new_folder: bool) -> "TrainingSpec":
        checkpoint = ZipUtil.unzip(path, "checkpoint.json", lambda x: json.load(open(x, "r")))
        config = ZipUtil.unzip(path, checkpoint["files"]["config"], lambda x: json.load(open(x, "r")))
        spec = TrainingSpec.from_conf(config, None, new_folder)
        spec.optimizer.state_path = checkpoint["files"]["optimizer"]
        spec.lora_layers.state_path = checkpoint["files"]["lora_adapter"]
        spec.training_loop.iterator_state_path = checkpoint["files"]["iterator"]
        spec.statistics.state_path = checkpoint["files"]["loss"]
        spec.checkpoint_path = path
        spec.config_path = None
        return spec



import math
from typing import TYPE_CHECKING

import mlx.core as mx

if TYPE_CHECKING:
    from mflux.models.common.config.config import Config

from mflux.models.common.schedulers.base_scheduler import BaseScheduler


class FlowMatchEulerDiscreteScheduler(BaseScheduler):
    def __init__(self, config: "Config"):
        self.config = config
        self.model_config = config.model_config
        self.num_train_timesteps = 1000
        self.shift_terminal = 0.02
        self.base_shift = 0.5
        self.max_shift = 0.9
        self.base_image_seq_len = 256
        self.max_image_seq_len = 8192
        self._sigmas, self._timesteps = self._compute_timesteps_and_sigmas()

    @property
    def sigmas(self) -> mx.array:
        return self._sigmas

    @property
    def timesteps(self) -> mx.array:
        return self._timesteps

    def _compute_mu(self) -> float:
        h_patches = self.config.height // 16
        w_patches = self.config.width // 16
        seq_len = h_patches * w_patches
        m = (self.max_shift - self.base_shift) / (self.max_image_seq_len - self.base_image_seq_len)
        b = self.base_shift - m * self.base_image_seq_len
        mu = m * seq_len + b
        return mu

    @staticmethod
    def _time_shift_exponential(mu: float, sigma_power: float, t: float) -> float:
        return math.exp(mu) / (math.exp(mu) + ((1.0 / t - 1.0) ** sigma_power))

    def _stretch_to_terminal(self, sigmas: list[float]) -> list[float]:
        one_minus_sigmas = [1.0 - s for s in sigmas]
        scale_factor = one_minus_sigmas[-1] / (1.0 - self.shift_terminal)
        stretched = [1.0 - (oms / scale_factor) for oms in one_minus_sigmas]
        return stretched

    def _compute_timesteps_and_sigmas(self) -> tuple[mx.array, mx.array]:
        num_steps = self.config.num_inference_steps
        sigma_min = 1.0 / self.num_train_timesteps
        sigma_max = 1.0
        timesteps_linear = [
            sigma_max * self.num_train_timesteps
            - i * (sigma_max - sigma_min) * self.num_train_timesteps / (num_steps - 1)
            for i in range(num_steps)
        ]
        sigmas_linear = [t / self.num_train_timesteps for t in timesteps_linear]
        sigmas_shifted = [FlowMatchEulerDiscreteScheduler._time_shift_exponential(1.0, 1.0, s) for s in sigmas_linear]
        sigmas_final = self._stretch_to_terminal(sigmas_shifted)
        timesteps = [s * self.num_train_timesteps for s in sigmas_final]
        sigmas_with_zero = sigmas_final + [0.0]
        sigmas_arr = mx.array(sigmas_with_zero, dtype=mx.float32)
        timesteps_arr = mx.array(timesteps, dtype=mx.float32)
        return sigmas_arr, timesteps_arr

    def step(self, noise: mx.array, timestep: int, latents: mx.array, **kwargs) -> mx.array:
        dt = self._sigmas[timestep + 1] - self._sigmas[timestep]
        return latents + dt * noise

    def scale_model_input(self, latents: mx.array, t: int) -> mx.array:
        return latents



import os
from pathlib import Path

import mlx.core as mx
import PIL.Image

from mflux.callbacks.callback import BeforeLoopCallback
from mflux.models.common.config.config import Config
from mflux.utils.image_util import ImageUtil


class DepthImageSaver(BeforeLoopCallback):
    def __init__(self, path: str):
        self.path = Path(path)

    def call_before_loop(
        self,
        seed: int,
        prompt: str,
        latents: mx.array,
        config: Config,
        canny_image: PIL.Image.Image | None = None,
        depth_image: PIL.Image.Image | None = None,
    ) -> None:
        if depth_image is None:
            return

        base, ext = os.path.splitext(self.path)
        ImageUtil.save_image(
            image=depth_image,
            path=f"{base}_depth_map{ext}",
        )



import mlx.core as mx
from mlx import nn

from mflux.models.z_image.model.z_image_vae.common.resnet_block_2d import ResnetBlock2D
from mflux.models.z_image.model.z_image_vae.decoder.up_sampler import UpSampler


class UpDecoderBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_layers: int = 3,
        add_upsample: bool = True,
    ):
        super().__init__()
        self.add_upsample = add_upsample

        # Build ResNet blocks
        resnets = []
        for i in range(num_layers):
            input_ch = in_channels if i == 0 else out_channels
            use_shortcut = (i == 0) and (in_channels != out_channels)
            resnets.append(
                ResnetBlock2D(
                    in_channels=input_ch,
                    out_channels=out_channels,
                    use_conv_shortcut=use_shortcut,
                )
            )
        self.resnets = resnets

        # Upsampler
        if add_upsample:
            self.upsamplers = [UpSampler(in_channels=out_channels, out_channels=out_channels)]
        else:
            self.upsamplers = None

    def __call__(self, hidden_states: mx.array) -> mx.array:
        for resnet in self.resnets:
            hidden_states = resnet(hidden_states)

        if self.upsamplers is not None:
            for upsampler in self.upsamplers:
                hidden_states = upsampler(hidden_states)

        return hidden_states



import mlx.core as mx

from mflux.models.common.tokenizer import Tokenizer
from mflux.models.qwen.model.qwen_text_encoder.qwen_text_encoder import QwenTextEncoder


class QwenPromptEncoder:
    @staticmethod
    def encode_prompt(
        prompt: str,
        negative_prompt: str,
        prompt_cache: dict[str, tuple[mx.array, mx.array, mx.array, mx.array]],
        qwen_tokenizer: Tokenizer,
        qwen_text_encoder: QwenTextEncoder,
    ) -> tuple[mx.array, mx.array, mx.array, mx.array]:
        # Use a space as fallback for empty negative prompt to ensure valid tokenization
        if not negative_prompt or not negative_prompt.strip():
            negative_prompt = " "

        # 0. Create a cache key that combines both prompts
        cache_key = f"{prompt}|NEG|{negative_prompt}"

        # 1. Return prompt encodings if already cached
        if cache_key in prompt_cache:
            return prompt_cache[cache_key]

        # 2. Encode the positive prompt
        pos_output = qwen_tokenizer.tokenize(prompt)
        prompt_embeds, prompt_mask = qwen_text_encoder(
            input_ids=pos_output.input_ids, attention_mask=pos_output.attention_mask
        )

        # 3. Encode the negative prompt
        neg_output = qwen_tokenizer.tokenize(negative_prompt)
        neg_prompt_embeds, neg_prompt_mask = qwen_text_encoder(
            input_ids=neg_output.input_ids, attention_mask=neg_output.attention_mask
        )

        # 4. Cache the result (all 4 values)
        result = (prompt_embeds, prompt_mask, neg_prompt_embeds, neg_prompt_mask)
        prompt_cache[cache_key] = result
        return result



from pathlib import Path
from typing import TYPE_CHECKING

import transformers
from huggingface_hub import snapshot_download

from mflux.models.common.tokenizer.tokenizer import (
    BaseTokenizer,
    LanguageTokenizer,
    VisionLanguageTokenizer,
)

if TYPE_CHECKING:
    from mflux.models.common.weights.loading.weight_definition import TokenizerDefinition


class TokenizerLoader:
    @staticmethod
    def load(
        definition: "TokenizerDefinition",
        model_path: str,
    ) -> BaseTokenizer:
        tokenizer_path = TokenizerLoader._resolve_path(
            model_path=model_path,
            hf_subdir=definition.hf_subdir,
            fallback_subdirs=definition.fallback_subdirs,
            download_patterns=definition.download_patterns,
        )

        raw_tokenizer = TokenizerLoader._load_raw_tokenizer(
            tokenizer_path=tokenizer_path,
            tokenizer_class=definition.tokenizer_class,
            chat_template=definition.chat_template,
        )

        return TokenizerLoader._create_tokenizer(
            raw_tokenizer=raw_tokenizer,
            definition=definition,
        )

    @staticmethod
    def load_all(
        definitions: list["TokenizerDefinition"],
        model_path: str,
        max_length_overrides: dict[str, int] | None = None,
    ) -> dict[str, BaseTokenizer]:
        max_length_overrides = max_length_overrides or {}
        result = {}
        for d in definitions:
            tokenizer = TokenizerLoader.load(d, model_path)
            if d.name in max_length_overrides:
                tokenizer.max_length = max_length_overrides[d.name]
            result[d.name] = tokenizer
        return result

    @staticmethod
    def _resolve_path(
        model_path: str,
        hf_subdir: str,
        fallback_subdirs: list[str] | None,
        download_patterns: list[str] | None,
    ) -> Path:
        expanded = Path(model_path).expanduser()
        if expanded.exists():
            root_path = expanded
        elif "/" in model_path and model_path.count("/") == 1 and not model_path.startswith(("./", "../")):
            patterns = download_patterns or [f"{hf_subdir}/**"]
            root_path = Path(
                snapshot_download(
                    repo_id=model_path,
                    allow_patterns=patterns,
                )
            )
        else:
            raise FileNotFoundError(
                f"Model not found: '{model_path}'. "
                f"If local path, make sure it exists. "
                f"If HuggingFace repo, use 'org/model' format."
            )

        tokenizer_path = root_path / hf_subdir
        if tokenizer_path.exists():
            return tokenizer_path

        if fallback_subdirs:
            for subdir in fallback_subdirs:
                if subdir == ".":
                    if TokenizerLoader._has_tokenizer_files(root_path):
                        return root_path
                else:
                    fallback_path = root_path / subdir
                    if fallback_path.exists():
                        return fallback_path

        return tokenizer_path

    @staticmethod
    def _has_tokenizer_files(path: Path) -> bool:
        tokenizer_indicators = ["vocab.json", "tokenizer.json", "tokenizer_config.json"]
        return any((path / f).exists() for f in tokenizer_indicators)

    @staticmethod
    def _load_raw_tokenizer(
        tokenizer_path: Path,
        tokenizer_class: str,
        chat_template: str | None = None,
    ):
        if hasattr(transformers, tokenizer_class):
            cls = getattr(transformers, tokenizer_class)
        else:
            raise ValueError(f"Unknown tokenizer class: {tokenizer_class}")

        # Workaround for Qwen2Tokenizer bug in transformers 5.0.0rc0
        # The tokenizer doesn't properly load vocab/merges from files, need to pass them directly
        if tokenizer_class == "Qwen2Tokenizer":
            tokenizer = TokenizerLoader._load_qwen2_tokenizer_workaround(tokenizer_path, cls)
        else:
            tokenizer = cls.from_pretrained(
                pretrained_model_name_or_path=str(tokenizer_path),
                local_files_only=True,
            )

        # Apply chat_template from definition if provided and not already set
        if chat_template and not getattr(tokenizer, "chat_template", None):
            tokenizer.chat_template = chat_template

        return tokenizer

    @staticmethod
    def _load_qwen2_tokenizer_workaround(tokenizer_path: Path, cls):
        """Load Qwen2Tokenizer with explicit vocab and merges data.

        In transformers 5.0, Qwen2Tokenizer builds its internal BPE tokenizer from the
        `vocab` and `merges` parameters BEFORE calling super().__init__(). The `vocab_file`
        and `merges_file` parameters are only passed to the parent class for saving purposes
        and are NOT automatically loaded. We must load and parse these files ourselves.

        We also need to load special tokens from tokenizer_config.json to ensure tokens like
        <|im_start|>, <|im_end|>, <|vision_start|>, etc. are properly registered.
        """
        import json

        from tokenizers import AddedToken

        vocab_file = tokenizer_path / "vocab.json"
        merges_file = tokenizer_path / "merges.txt"
        config_file = tokenizer_path / "tokenizer_config.json"

        if not vocab_file.exists() or not merges_file.exists():
            # Fall back to standard loading if files don't exist
            return cls.from_pretrained(
                pretrained_model_name_or_path=str(tokenizer_path),
                local_files_only=True,
            )

        # Load vocab as dict[str, int]
        with open(vocab_file, encoding="utf-8") as f:
            vocab = json.load(f)

        # Load merges as list of tuples (pair of strings)
        # The merges.txt file has format: "token1 token2" per line (skip header if present)
        merges = []
        with open(merges_file, encoding="utf-8") as f:
            for line in f:
                line = line.rstrip()
                # Skip empty lines and the optional version header (e.g., "#version: 0.2")
                if not line or line.startswith("#"):
                    continue
                parts = line.split()
                if len(parts) == 2:
                    merges.append((parts[0], parts[1]))

        # Load tokenizer config for special tokens and chat template
        config_kwargs = {}
        chat_template = None
        if config_file.exists():
            with open(config_file, encoding="utf-8") as f:
                config = json.load(f)

            # Extract added_tokens_decoder to add special tokens
            added_tokens_decoder = config.get("added_tokens_decoder", {})
            if added_tokens_decoder:
                # Convert to the format expected by the tokenizer
                config_kwargs["added_tokens_decoder"] = {
                    int(k): AddedToken(
                        content=v["content"],
                        lstrip=v.get("lstrip", False),
                        rstrip=v.get("rstrip", False),
                        single_word=v.get("single_word", False),
                        normalized=v.get("normalized", False),
                        special=v.get("special", True),
                    )
                    for k, v in added_tokens_decoder.items()
                }

            # Extract chat_template if present
            chat_template = config.get("chat_template")

        tokenizer = cls(vocab=vocab, merges=merges, **config_kwargs)

        # Set chat_template after initialization (not a constructor param)
        if chat_template:
            tokenizer.chat_template = chat_template

        return tokenizer

    @staticmethod
    def _create_tokenizer(
        raw_tokenizer,
        definition: "TokenizerDefinition",
    ) -> BaseTokenizer:
        encoder_class = definition.encoder_class

        if encoder_class is VisionLanguageTokenizer:
            if definition.processor_class is None:
                raise ValueError("VisionLanguageTokenizer requires processor_class in definition")
            processor = definition.processor_class(tokenizer=raw_tokenizer)
            return VisionLanguageTokenizer(
                tokenizer=raw_tokenizer,
                processor=processor,
                max_length=definition.max_length,
                template=definition.template,
                image_token=definition.image_token,
            )
        else:
            # Default to LanguageTokenizer for all text-only cases
            return LanguageTokenizer(
                tokenizer=raw_tokenizer,
                max_length=definition.max_length,
                padding=definition.padding,
                template=definition.template,
                use_chat_template=definition.use_chat_template,
                chat_template_kwargs=definition.chat_template_kwargs or {},
                add_special_tokens=definition.add_special_tokens,
            )



import mlx.core as mx
from mlx import nn
from mlx.core.fast import scaled_dot_product_attention

from mflux.models.common.config import ModelConfig


class Attention(nn.Module):
    def __init__(self, channels: int = 512, num_groups: int = 32):
        super().__init__()
        self.channels = channels

        self.group_norm = nn.GroupNorm(
            num_groups=num_groups,
            dims=channels,
            eps=1e-6,
            affine=True,
            pytorch_compatible=True,
        )
        self.to_q = nn.Linear(channels, channels)
        self.to_k = nn.Linear(channels, channels)
        self.to_v = nn.Linear(channels, channels)
        self.to_out = [nn.Linear(channels, channels)]

    def __call__(self, input_array: mx.array) -> mx.array:
        input_array = mx.transpose(input_array, (0, 2, 3, 1))
        B, H, W, C = input_array.shape

        # Group norm
        hidden_states = self.group_norm(input_array.astype(mx.float32)).astype(ModelConfig.precision)

        # QKV projections - reshape for attention
        queries = self.to_q(hidden_states).reshape(B, H * W, 1, C)
        keys = self.to_k(hidden_states).reshape(B, H * W, 1, C)
        values = self.to_v(hidden_states).reshape(B, H * W, 1, C)

        # Transpose to (B, num_heads, seq_len, head_dim)
        queries = mx.transpose(queries, (0, 2, 1, 3))
        keys = mx.transpose(keys, (0, 2, 1, 3))
        values = mx.transpose(values, (0, 2, 1, 3))

        # Scaled dot product attention
        scale = 1.0 / mx.sqrt(mx.array(queries.shape[-1], dtype=queries.dtype))
        hidden_states = scaled_dot_product_attention(queries, keys, values, scale=scale)
        hidden_states = mx.transpose(hidden_states, (0, 2, 1, 3)).reshape(B, H, W, C)
        hidden_states = self.to_out[0](hidden_states)
        output = input_array + hidden_states
        return mx.transpose(output, (0, 3, 1, 2))



import mlx.core as mx
from mlx import nn


class SmolLM3_3B_RMSNorm(nn.Module):
    def __init__(self, hidden_size: int, eps: float = 1e-6):
        super().__init__()
        self.weight = mx.ones((hidden_size,))
        self.eps = eps

    def __call__(self, hidden_states: mx.array) -> mx.array:
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.astype(mx.float32)
        variance = mx.mean(mx.square(hidden_states), axis=-1, keepdims=True)
        hidden_states = hidden_states * mx.rsqrt(variance + self.eps)
        result = self.weight.astype(mx.float32) * hidden_states
        return result.astype(input_dtype)



import json
import logging
import re
import subprocess

from mflux.callbacks.callback import BeforeLoopCallback
from mflux.utils.exceptions import StopImageGenerationException

logger = logging.getLogger(__name__)


class BatterySaver(BeforeLoopCallback):
    PMSET_AC_POWER_STATUS = "Now drawing from 'AC Power'"
    PMSET_BATT_STATUS_PATTERN = r"InternalBattery-.+?(\d+)%"

    _machine_model: str | None = None
    _is_battery_powered: bool | None = None

    def __init__(self, battery_percentage_stop_limit: int = 10):
        self.limit = battery_percentage_stop_limit

    def call_before_loop(self, **kwargs) -> None:  # type: ignore
        current_pct = self._get_battery_percentage()
        if current_pct is not None and current_pct <= self.limit:
            raise StopImageGenerationException(f"Battery below {self.limit}% threshold: {current_pct}%")

    def _get_battery_percentage(self) -> int | None:
        if not self._is_machine_battery_powered():
            return None

        percentage = None
        try:
            result = subprocess.run(
                ["pmset", "-g", "batt"],
                capture_output=True,
                text=True,
                check=True,
            )
            if self.PMSET_AC_POWER_STATUS not in result.stdout:
                if match := re.search(self.PMSET_BATT_STATUS_PATTERN, result.stdout):
                    percentage = int(match.group(1))
        except (subprocess.CalledProcessError, TypeError) as e:
            logger.warning(
                f"Cannot read battery percentage via 'pmset -g batt': {e}. "
                f"Battery saver functionality is disabled and the program will continue running."
            )

        return percentage

    @classmethod
    def _is_machine_battery_powered(cls) -> bool:
        if cls._is_battery_powered is None:
            machine_model = cls._get_machine_model()
            cls._is_battery_powered = "MacBook" in machine_model
        return cls._is_battery_powered

    @classmethod
    def _get_machine_model(cls) -> str:
        if cls._machine_model is None:
            try:
                result = subprocess.run(
                    ["system_profiler", "-json", "SPHardwareDataType"],
                    capture_output=True,
                    text=True,
                    check=True,
                )
                data = json.loads(result.stdout)
                cls._machine_model = data["SPHardwareDataType"][0]["machine_model"]
            except (subprocess.CalledProcessError, json.JSONDecodeError, IndexError, KeyError) as e:
                logger.warning(f"Cannot determine machine model via 'system_profiler -json SPHardwareDataType': {e}")
                cls._machine_model = "Unknown"
        return cls._machine_model



from pathlib import Path

import mlx.core as mx
from mlx import nn

from mflux.models.common.config import ModelConfig
from mflux.models.common.config.config import Config
from mflux.models.common.latent_creator.latent_creator import Img2Img, LatentCreator
from mflux.models.common.vae.vae_util import VAEUtil
from mflux.models.common.weights.saving.model_saver import ModelSaver
from mflux.models.qwen.latent_creator.qwen_latent_creator import QwenLatentCreator
from mflux.models.qwen.model.qwen_text_encoder.qwen_prompt_encoder import QwenPromptEncoder
from mflux.models.qwen.model.qwen_text_encoder.qwen_text_encoder import QwenTextEncoder
from mflux.models.qwen.model.qwen_transformer.qwen_transformer import QwenTransformer
from mflux.models.qwen.model.qwen_vae.qwen_vae import QwenVAE
from mflux.models.qwen.qwen_initializer import QwenImageInitializer
from mflux.models.qwen.weights.qwen_weight_definition import QwenWeightDefinition
from mflux.utils.exceptions import StopImageGenerationException
from mflux.utils.generated_image import GeneratedImage
from mflux.utils.image_util import ImageUtil


class QwenImage(nn.Module):
    vae: QwenVAE
    transformer: QwenTransformer
    text_encoder: QwenTextEncoder

    def __init__(
        self,
        quantize: int | None = None,
        model_path: str | None = None,
        lora_paths: list[str] | None = None,
        lora_scales: list[float] | None = None,
        model_config: ModelConfig = ModelConfig.qwen_image(),
    ):
        super().__init__()
        QwenImageInitializer.init(
            model=self,
            quantize=quantize,
            model_path=model_path,
            lora_paths=lora_paths,
            lora_scales=lora_scales,
            model_config=model_config,
        )

    def generate_image(
        self,
        seed: int,
        prompt: str,
        num_inference_steps: int = 4,
        height: int = 1024,
        width: int = 1024,
        guidance: float = 4.0,
        image_path: Path | str | None = None,
        image_strength: float | None = None,
        scheduler: str = "linear",
        negative_prompt: str | None = None,
    ) -> GeneratedImage:
        # 0. Create a new config based on the model type and input parameters
        config = Config(
            width=width,
            height=height,
            guidance=guidance,
            scheduler=scheduler,
            image_path=image_path,
            image_strength=image_strength,
            model_config=self.model_config,
            num_inference_steps=num_inference_steps,
        )

        # 1. Create the initial latents
        latents = LatentCreator.create_for_txt2img_or_img2img(
            seed=seed,
            width=config.width,
            height=config.height,
            img2img=Img2Img(
                vae=self.vae,
                latent_creator=QwenLatentCreator,
                sigmas=config.scheduler.sigmas,
                init_time_step=config.init_time_step,
                image_path=config.image_path,
                tiling_config=self.tiling_config,
            ),
        )

        # 2. Encode the prompt (using native MLX encoding)
        prompt_embeds, prompt_mask, negative_prompt_embeds, negative_prompt_mask = QwenPromptEncoder.encode_prompt(
            prompt=prompt,
            negative_prompt=negative_prompt,
            prompt_cache=self.prompt_cache,
            qwen_tokenizer=self.tokenizers["qwen"],
            qwen_text_encoder=self.text_encoder,
        )

        # 3. Create callback context and call before_loop
        ctx = self.callbacks.start(seed=seed, prompt=prompt, config=config)
        ctx.before_loop(latents)

        for t in config.time_steps:
            try:
                # Scale model input if needed by the scheduler
                latents = config.scheduler.scale_model_input(latents, t)

                # 4. Predict the noise
                noise = self.transformer(
                    t=t,
                    config=config,
                    hidden_states=latents,
                    encoder_hidden_states=prompt_embeds,
                    encoder_hidden_states_mask=prompt_mask,
                )
                noise_negative = self.transformer(
                    t=t,
                    config=config,
                    hidden_states=latents,
                    encoder_hidden_states=negative_prompt_embeds,
                    encoder_hidden_states_mask=negative_prompt_mask,
                )
                guided_noise = QwenImage.compute_guided_noise(noise, noise_negative, config.guidance)

                # 5.t Take one denoise step
                latents = config.scheduler.step(noise=guided_noise, timestep=t, latents=latents)

                # 6.t Call subscribers in-loop
                ctx.in_loop(t, latents)

                # (Optional) Evaluate to enable progress tracking
                mx.eval(latents)

            except KeyboardInterrupt:  # noqa: PERF203
                ctx.interruption(t, latents)
                raise StopImageGenerationException(
                    f"Stopping image generation at step {t + 1}/{config.num_inference_steps}"
                )

        # 7. Call subscribers after loop
        ctx.after_loop(latents)

        # 8. Decode the latent array and return the image
        latents = QwenLatentCreator.unpack_latents(latents=latents, height=config.height, width=config.width)
        decoded = VAEUtil.decode(vae=self.vae, latent=latents, tiling_config=self.tiling_config)
        return ImageUtil.to_image(
            decoded_latents=decoded,
            config=config,
            seed=seed,
            prompt=prompt,
            quantization=self.bits,
            lora_paths=self.lora_paths,
            lora_scales=self.lora_scales,
            image_path=config.image_path,
            image_strength=config.image_strength,
            generation_time=config.time_steps.format_dict["elapsed"],
            negative_prompt=negative_prompt,
        )

    def save_model(self, base_path: str) -> None:
        ModelSaver.save_model(
            model=self,
            bits=self.bits,
            base_path=base_path,
            weight_definition=QwenWeightDefinition,
        )

    @staticmethod
    def compute_guided_noise(
        noise: mx.array,
        noise_negative: mx.array,
        guidance: float,
    ) -> mx.array:
        combined = noise_negative + guidance * (noise - noise_negative)
        cond_norm = mx.sqrt(mx.sum(noise * noise, axis=-1, keepdims=True) + 1e-12)
        noise_norm = mx.sqrt(mx.sum(combined * combined, axis=-1, keepdims=True) + 1e-12)
        noise = combined * (cond_norm / noise_norm)
        return noise



import mlx.core as mx
from mlx import nn

from mflux.models.seedvr2.model.seedvr2_transformer.ada_modulation import AdaModulation
from mflux.models.seedvr2.model.seedvr2_transformer.attention import MMAttention
from mflux.models.seedvr2.model.seedvr2_transformer.mm_swiglu import MMSwiGLU


class TransformerBlock(nn.Module):
    def __init__(
        self,
        vid_dim: int = 2560,
        txt_dim: int = 2560,
        heads: int = 20,
        head_dim: int = 128,
        expand_ratio: int = 4,
        norm_eps: float = 1e-5,
        qk_bias: bool = False,
        rope_dim: int = 128,
        shared_weights: bool = False,
        is_last_layer: bool = False,
        window: tuple[int, int, int] = (4, 3, 3),
        shift: bool = False,
    ):
        super().__init__()
        self.shared_weights = shared_weights
        self.is_last_layer = is_last_layer
        self.vid_dim = vid_dim
        self.txt_dim = txt_dim
        self.norm_eps = norm_eps

        self.attn = MMAttention(
            vid_dim=vid_dim,
            txt_dim=txt_dim,
            heads=heads,
            head_dim=head_dim,
            qk_bias=qk_bias,
            qk_norm_eps=norm_eps,
            rope_dim=rope_dim,
            shared_weights=shared_weights,
            window=window,
            shift=shift,
        )

        self.mlp = MMSwiGLU(
            vid_dim=vid_dim,
            txt_dim=txt_dim,
            expand_ratio=expand_ratio,
            shared_weights=shared_weights,
            is_last_layer=is_last_layer,
        )

        self.ada = AdaModulation(
            dim=vid_dim,
            shared_weights=shared_weights,
            is_last_layer=is_last_layer,
        )

    def __call__(
        self,
        vid: mx.array,
        txt: mx.array,
        emb: mx.array,
        vid_shape: mx.array,
        txt_shape: mx.array,
    ) -> tuple[mx.array, mx.array]:
        vid_attn = TransformerBlock._rms_norm(vid, self.norm_eps)
        txt_attn = TransformerBlock._rms_norm(txt, self.norm_eps)

        vid_attn = self.ada.modulate_vid(vid_attn, emb, layer="attn", mode="in")
        txt_attn = self.ada.modulate_txt(txt_attn, emb, layer="attn", mode="in")

        vid_attn, txt_attn = self.attn(vid_attn, txt_attn, vid_shape, txt_shape)

        vid_attn = self.ada.modulate_vid(vid_attn, emb, layer="attn", mode="out")
        txt_attn = self.ada.modulate_txt(txt_attn, emb, layer="attn", mode="out")

        vid = vid + vid_attn
        if not self.is_last_layer:
            txt = txt + txt_attn

        vid_mlp = TransformerBlock._rms_norm(vid, self.norm_eps)
        if self.is_last_layer:
            txt_mlp = txt
        else:
            txt_mlp = TransformerBlock._rms_norm(txt, self.norm_eps)

        vid_mlp = self.ada.modulate_vid(vid_mlp, emb, layer="mlp", mode="in")
        txt_mlp = self.ada.modulate_txt(txt_mlp, emb, layer="mlp", mode="in")

        vid_mlp, txt_mlp = self.mlp(vid_mlp, txt_mlp)

        vid_mlp = self.ada.modulate_vid(vid_mlp, emb, layer="mlp", mode="out")
        txt_mlp = self.ada.modulate_txt(txt_mlp, emb, layer="mlp", mode="out")

        vid = vid + vid_mlp
        if not self.is_last_layer:
            txt = txt + txt_mlp

        return vid, txt

    @staticmethod
    def _rms_norm(x: mx.array, eps: float = 1e-5) -> mx.array:
        return mx.fast.rms_norm(x, mx.ones(x.shape[-1]), eps)



import mlx.core as mx
from mlx import nn

from mflux.models.flux.model.flux_vae.decoder.decoder import Decoder
from mflux.models.flux.model.flux_vae.encoder.encoder import Encoder


class VAE(nn.Module):
    scaling_factor: int = 0.3611
    shift_factor: int = 0.1159
    spatial_scale = 8
    latent_channels = 16

    def __init__(self):
        super().__init__()
        self.decoder = Decoder()
        self.encoder = Encoder()

    def decode(self, latents: mx.array) -> mx.array:
        if latents.ndim == 5:
            latents = latents[:, :, 0, :, :]
        scaled_latents = (latents / self.scaling_factor) + self.shift_factor
        decoded = self.decoder(scaled_latents)
        return decoded[:, :, None, :, :]

    def encode(self, image: mx.array) -> mx.array:
        if image.ndim == 5:
            image = image[:, :, 0, :, :]
        latents = self.encoder(image)
        mean, _ = mx.split(latents, 2, axis=1)
        latent = (mean - self.shift_factor) * self.scaling_factor
        return latent[:, :, None, :, :]



import mlx.core as mx
from mlx import nn


class ReduxEncoder(nn.Module):
    def __init__(self):
        super().__init__()
        self.redux_up = nn.Linear(1152, 4096 * 3)
        self.redux_down = nn.Linear(4096 * 3, 4096)

    def __call__(self, x: mx.array) -> mx.array:
        return self.redux_down(nn.silu(self.redux_up(x)))



import mlx.core as mx
from mlx import nn


class QwenRotaryEmbedding(nn.Module):
    def __init__(
        self,
        dim: int,
        max_position_embeddings: int = 128000,
        base: float = 1000000.0,
        device: str = None,
        scaling_factor: float = 1.0,
        rope_type: str = "default",
        config=None,
    ):
        super().__init__()
        self.inv_freq = 1.0 / (base ** (mx.arange(0, dim, 2, dtype=mx.float32) / dim))
        self.attention_scaling = scaling_factor

    def __call__(self, x: mx.array, position_ids: mx.array) -> tuple[mx.array, mx.array]:
        if len(position_ids.shape) == 2:
            batch_size, seq_len = position_ids.shape
            position_ids = mx.broadcast_to(mx.expand_dims(position_ids, axis=0), (3, batch_size, seq_len))

        inv_freq_expanded = mx.expand_dims(mx.expand_dims(self.inv_freq, axis=0), axis=0)
        inv_freq_expanded = mx.expand_dims(inv_freq_expanded, axis=-1)
        inv_freq_expanded = mx.broadcast_to(inv_freq_expanded, (3, position_ids.shape[1], self.inv_freq.shape[0], 1))
        inv_freq_expanded = inv_freq_expanded.astype(mx.float32)
        position_ids_expanded = mx.expand_dims(position_ids, axis=2)
        position_ids_expanded = position_ids_expanded.astype(mx.float32)
        freqs = mx.matmul(inv_freq_expanded, position_ids_expanded)
        freqs = mx.transpose(freqs, (0, 1, 3, 2))
        emb = mx.concatenate([freqs, freqs], axis=-1)
        cos = mx.cos(emb) * self.attention_scaling
        sin = mx.sin(emb) * self.attention_scaling
        return cos.astype(x.dtype), sin.astype(x.dtype)



from pathlib import Path

import mlx.core as mx

from mflux.models.flux.model.redux_encoder.redux_encoder import ReduxEncoder
from mflux.models.flux.model.siglip_vision_transformer.siglip_vision_transformer import SiglipVisionTransformer
from mflux.utils.image_util import ImageUtil


class ReduxUtil:
    @staticmethod
    def validate_redux_image_strengths(
        redux_image_paths: list[Path],
        redux_image_strengths: list[float] | None,
    ) -> list[float] | None:
        if not redux_image_strengths or len(redux_image_strengths) == 0:
            return redux_image_strengths

        # If strengths provided but not enough for all images, fill with default (1.0)
        if len(redux_image_strengths) < len(redux_image_paths):
            redux_image_strengths.extend([1.0] * (len(redux_image_paths) - len(redux_image_strengths)))

        # If too many strengths provided, raise error
        elif len(redux_image_strengths) > len(redux_image_paths):
            raise ValueError(
                f"Too many strengths provided ({len(redux_image_strengths)}), expted at most {len(redux_image_paths)}."
            )

        return redux_image_strengths

    @staticmethod
    def embed_images(
        image_paths: list[str] | list[Path],
        image_encoder: SiglipVisionTransformer,
        image_embedder: ReduxEncoder,
        image_strengths: list[float] | None = None,
    ) -> list[mx.array]:  # fmt:off
        image_embeds_list = []
        for idx, image_path in enumerate(image_paths):
            # Get the strength for this image (default to 1.0 if not specified)
            strength = 1.0
            if image_strengths is not None and idx < len(image_strengths):
                strength = image_strengths[idx]

            image_embeds = ReduxUtil._embed_single_image(
                image_path=image_path,
                image_encoder=image_encoder,
                image_embedder=image_embedder,
                strength=strength,
            )
            image_embeds_list.append(image_embeds)
        return image_embeds_list

    @staticmethod
    def _embed_single_image(
        image_path: str | Path,
        image_encoder: SiglipVisionTransformer,
        image_embedder: ReduxEncoder,
        strength: float = 1.0,
    ) -> mx.array:  # fmt:off
        image = ImageUtil.load_image(image_path).convert("RGB")
        image = ImageUtil.preprocess_for_model(image=image)
        image_latents, pooler_output = image_encoder(image)
        image_embeds = image_embedder(image_latents)

        # Apply strength factor to the image embeddings
        if strength != 1.0:
            image_embeds = image_embeds * strength

        return image_embeds



import logging
from pathlib import Path

import mlx.core as mx
from tqdm import tqdm

from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.schedulers import SCHEDULER_REGISTRY, try_import_external_scheduler
from mflux.models.common.schedulers.linear_scheduler import LinearScheduler

logger = logging.getLogger(__name__)


class Config:
    def __init__(
        self,
        model_config: ModelConfig,
        num_inference_steps: int = 4,
        height: int = 1024,
        width: int = 1024,
        guidance: float = 4.0,
        image_path: Path | str | None = None,
        image_strength: float | None = None,
        depth_image_path: Path | str | None = None,
        redux_image_paths: list[Path | str] | None = None,
        redux_image_strengths: list[float] | None = None,
        masked_image_path: Path | str | None = None,
        controlnet_strength: float | None = None,
        scheduler: str = "linear",
    ):
        # Ensure dimensions are multiples of 16
        if width % 16 != 0 or height % 16 != 0:
            logger.warning("Width and height should be multiples of 16. Rounding down.")

        self.model_config = model_config
        self._num_inference_steps = num_inference_steps
        self._height = 16 * (height // 16)
        self._width = 16 * (width // 16)
        self._guidance = guidance
        self._image_path = Path(image_path) if isinstance(image_path, str) else image_path
        self._image_strength = image_strength
        self._depth_image_path = Path(depth_image_path) if isinstance(depth_image_path, str) else depth_image_path
        self._redux_image_paths = (
            [Path(p) if isinstance(p, str) else p for p in redux_image_paths] if redux_image_paths else None
        )
        self._redux_image_strengths = redux_image_strengths
        self._masked_image_path = Path(masked_image_path) if isinstance(masked_image_path, str) else masked_image_path
        self._controlnet_strength = controlnet_strength
        self._scheduler_str = scheduler
        self._scheduler = None
        self._time_steps = None

    @property
    def height(self) -> int:
        return self._height

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, value):
        self._width = value

    @property
    def guidance(self) -> float:
        return self._guidance

    @property
    def num_inference_steps(self) -> int:
        return self._num_inference_steps

    @property
    def precision(self) -> mx.Dtype:
        return ModelConfig.precision

    @property
    def num_train_steps(self) -> int:
        return self.model_config.num_train_steps

    @property
    def image_path(self) -> Path | None:
        return self._image_path

    @property
    def image_strength(self) -> float | None:
        return self._image_strength

    @property
    def depth_image_path(self) -> Path | None:
        return self._depth_image_path

    @property
    def redux_image_paths(self) -> list[Path] | None:
        return self._redux_image_paths

    @property
    def redux_image_strengths(self) -> list[float] | None:
        return self._redux_image_strengths

    @property
    def masked_image_path(self) -> Path | None:
        return self._masked_image_path

    @property
    def init_time_step(self) -> int:
        is_img2img = (
            self._image_path is not None and
            self._image_strength is not None and
            self._image_strength > 0.0
        )  # fmt: off

        if is_img2img:
            # 1. Clamp strength to [0, 1]
            strength = max(0.0, min(1.0, self._image_strength))  # type: ignore

            # 2. Return start time in [1, floor(num_steps * strength)]
            return max(1, int(self._num_inference_steps * strength))  # type: ignore
        else:
            return 0

    @property
    def time_steps(self) -> tqdm:
        if self._time_steps is None:
            self._time_steps = tqdm(range(self.init_time_step, self.num_inference_steps))
        return self._time_steps

    @property
    def controlnet_strength(self) -> float | None:
        return self._controlnet_strength

    @property
    def scheduler(self):
        if self._scheduler is not None:
            return self._scheduler

        if self._scheduler_str == "linear":
            self._scheduler = LinearScheduler(self)
        elif (registered_scheduler := SCHEDULER_REGISTRY.get(self._scheduler_str, None)) is not None:
            self._scheduler = registered_scheduler(self)
        elif "." in self._scheduler_str:
            # this raises ValueError if scheduler is not importable
            scheduler_cls = try_import_external_scheduler(self._scheduler_str)
            self._scheduler = scheduler_cls(self)
        else:
            raise NotImplementedError(f"The scheduler {self._scheduler_str!r} is not implemented by mflux.")

        return self._scheduler



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class TilingConfig:
    vae_decode_tiles_per_dim: int | None = 8
    vae_decode_overlap: int = 8
    vae_encode_tiled: bool = True
    vae_encode_tile_size: int = 512
    vae_encode_tile_overlap: int = 64
